; The AKern Kernel
;   by AXEstudios

; AKern(AXE's Kernel) is an IBM PS/2 compatible Operating System Kernel
;   designed to be fast but also easy to write software for

; Some technical info and other stuff:
;   AKern's bootloader leaves all of the setup to the kernel
;   The system stack pointer is located at the address 0x10ffff
;   Memory is segmented into two parts, System and General Memory
;   The System memory segment is 0x200000 bytes big
;   The General memory segment fills the rest of the available memory.
;   For now, I plan on AKern only having support for IDE storage drives

;   I will admit I did use some code snippets from the internet
;       for parts of this


[bits 16]
[org 0x7e01]

call enableA20      ; enable A20 line
;                       [Using code from "enableA20.s" by J. Andrew McLaughlin]
cli                 ; disable interrupts
lgdt [gdt_desc]     ; load the GDT
mov eax, cr0        ; switch to protected mode
or eax, 0x1
mov cr0, eax
jmp CSEG_1:MAIN32   ; do long jump into main protected mode task

jmp $

; External Includes
%include "enba20.asm"
%include "gdt.asm"
%include "disp.asm"
%include "sr.asm"
%include "mem.asm"
%include "diskio.asm"
%include "cpu_interrupts.asm"


[bits 32]
    MAIN32:
; Immediately required setup
mov ax, DSEG_1
mov ds, ax
mov ss, ax
mov es, ax
mov fs, ax
mov gs, ax
mov esp, 0x10ffff ; resetting the stack pointer


    ;;;;;;;;; Main Task


call cleardisp      ; Clear the VGA Display

mov ecx, 0          ; Set the Row to display Text from
call setrow
    ; Startup Welcome Message
mov eax, msg_hello
call strout

    ; Note Message
call incrow

mov eax, msg_note
mov bl, 0x07
call strout_color

mov eax, 2
call addrow


;##
; ## CPU testing and setup
;  ##

    ; print start message
call incrow
mov eax, setup0
call strout
    ; check for 32-bit cpu
pushf
		pushf
		pop bx		; old FLAGS -> BX
		mov ax,bx
		xor ah,70h	; try changing b14 (NT)...
		push ax		; ... or b13:b12 (IOPL)
		popf
		pushf
		pop ax		; new FLAGS -> AX
popf
xor ah,bh		; 32-bit CPU if we changed NT...
and ah,70h		; ...or IOPL
jne is_32bit_cpu
mov eax, err32
call strout
jmp $
is_32bit_cpu:
    ; check if CPUID instruction is supported
call IS_CPUID_SUPPORTED
cmp eax, 0
jz .cpuid_not_supported
    ; Get vendor id
mov eax, 0x0
cpuid
    ; Dump and print the vendor id
mov eax, vendor_id
mov [eax], ebx
add eax, 4
mov [eax], edx
add eax, 4
mov [eax], ecx

call incrow
mov eax, msg_vendor_id
mov bl, 0x07
call strout_color
mov eax, vendor_id
call strout

    ; Check the vendor id
mov eax, vendor_id
mov ebx, vid_expected
mov edx, 1
call cmp_str
cmp edx, 0
jz .vid_not_compatible

    ; Check for a local APIC
    ; * Code for this has not been added yet.

    ; No extra code is being added for detecting
    ; CPU info right now.

jmp .done_checking_cpu

.cpuid_not_supported:
call incrow
mov eax, not0
mov bl, 0x07
call strout_color
jmp .done_checking_cpu
.vid_not_compatible:
    ; Since I am not currently trying
    ; to utilize the APIC,
    ; this will just be skipped.
    ; It will only cause issues with non-intel processors
    ; that aren't practical right now.
jmp .done_checking_cpu
;;;
call incrow
mov eax, errvid
mov bl, 0x04
call strout_color
jmp $
.done_checking_cpu:


;##
; ## PIC setup
;  ##

    ; print start message
call incrow
mov eax, setup1
call strout
    ; initialize
mov al, 0x11
out 0x20, al
out 0xA0, al
    ; remap
mov al, 0x20
out 0x21, al
mov al, 0x28
out 0xA1, al

mov al, 4
out 0x21, al
mov al, 2
out 0xA1, al

mov al, 1
out 0x21, al
out 0xA1, al
    ; enable keyboard input
mov al, 0b11111001
out 0x21, al
    ; print done message
mov eax, msg_done
call strout


;##
; ## CPU interrupt setup
;  ##

call incrow
mov eax, setup2 ; Print start message
call strout

call idt_setup
call isr_setup
lidt [idtr]
sti

int 0x12
mov al, [INTSTAT]
cmp al, 20
je .intworks
mov eax, errint
mov bl, 0x04
call strout_color
jmp $
.intworks:

; print done message
mov eax, msg_done
call strout

; Testing memory management (for reference)
;call ce_restart
;mov eax, 4
;call ce_alloc_wrapped
;mov eax, [cur_empty]
;mov dword [eax], "Hi!"
;mov eax, 4
;call ce_alloc_wrapped
;mov ebx, 0

;call ce_freelast_wrapped
;mov eax, [cur_empty]
;call strout



;##
; ## Testing Disk I/O
;  ##


    ;; DiskIN
call incrow
mov eax, setup3 ; start message
call strout

mov ah, 0a0h
mov bl, 1
mov bh, 1
mov cl, 0
mov ch, 0
mov edx, 0
call diskin

mov eax, 255
mov ebx, diskin_expected
mov edx, 1
call cmp_str
cmp edx, 0
jz .diskreadfail
jmp .diskreaddone

.diskreadfail:
call incrow
mov eax, errdiskio
mov bl, 0x04
call strout_color
jmp $

.diskreaddone:
mov eax, msg_done
call strout


;jmp .hltloop

;call loads

;jmp jump_usermode

;##
; ## Setting Up the User-Space
;  ##


call incrow
mov eax, setup4
call strout

mov ecx, tss
mov eax, tss
mov edx, tss_end
sub edx, eax
call set_tss_desc
call incrow

mov eax, msg_tss_desc
mov bl, 0x07
call strout_color
call disp_tss_desc

call tss_set_stack
mov ax, 0x28
ltr ax
call loads
; (ignore this ->) jmp 0x200000
; Can we jump to usermode now?

jmp enter_usermode

.hltloop:
hlt
jmp .hltloop

vendor_id: times 13 db 0
msg_vendor_id: db "Vendor ID: ",0
msg_tss_desc: db "TSS Descriptor: ",0
vid_expected: db "GenuineIntel",0
diskin_expected: db "AKern disk read test!",0
cur_row: dd 0
msg_done: db "Done!",0
msg_hello: db "AKern succesfully loaded into memory!",0
msg_note: db "NOTE: AKern uses VGA text mode. Different computers may or may not support this.",0
errdiskio: db "ERROR: Disk read test failed!",0
err32: db "ERROR: non-32-bit Architecture Detected!",0
errint: db "ERROR: Interrupts don't seem to work!",0
errnf: db "Address not found in the Memory Block Index!",0
errvid: db "ERROR: Incompatible Vendor-ID!",0
not0: db "[NOTE] CPUID has either not been enabled or is not supported.",0
setup0: db "[SETUP] Checking the CPU... ",0
setup1: db "[SETUP] Initializing and setting the PIC... ",0
setup2: db "[SETUP] Setting up cpu interrupts... ",0
setup3: db "[SETUP] Testing IDE disk read/write routines... ",0
setup4: db "[SETUP] Setting up the User-Space... ",0
setup5: db "[SETUP] Loading & Executing the AutoStart program... ",0




times 9216-($-$$) db 0 ; make this 18 sectors big

; some test user-mode code
add esp, 4
int 0x13
int 0x13
;;testing the division error interrupt
;mov edx, 0        ; clear dividend
;mov eax, 0x8003   ; dividend
;mov ecx, 0;x100    ; divisor
;div ecx           ; EAX = 0x80, EDX = 0x3
jmp $
